from typing import Any, TypeVar, Type
from remerkleable.core import BasicView, View

V = TypeVar('V', bound=View)


class boolean(int, BasicView):

    def encode_bytes(self) -> bytes:
        return b"\x01" if self else b"\x00"

    def __new__(cls, value: int):  # int value, but can be any subclass of int (bool, Bit, Bool, etc...)
        if value < 0 or value > 1:
            raise ValueError(f"value {value} out of bounds for bit")
        return super().__new__(cls, value)

    def __bool__(self):
        return self > 0

    @classmethod
    def coerce_view(cls: Type[V], v: Any) -> V:
        return cls(v)

    @classmethod
    def type_byte_length(cls) -> int:
        return 1

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        return cls(bytez != b"\x00")

    @classmethod
    def type_repr(cls) -> str:
        return "boolean"


class uint(int, BasicView):
    def __new__(cls, value: int):
        if value < 0:
            raise ValueError(f"unsigned type {cls} must not be negative")
        byte_len = cls.type_byte_length()
        if value.bit_length() > (byte_len << 3):
            raise ValueError(f"value out of bounds for {cls}")
        return super().__new__(cls, value)

    def __add__(self, other):
        return self.__class__(super().__add__(self.__class__.coerce_view(other)))

    def __sub__(self, other):
        return self.__class__(super().__sub__(self.__class__.coerce_view(other)))

    @classmethod
    def coerce_view(cls: Type[V], v: Any) -> V:
        if isinstance(v, uint) and cls.type_byte_length() != v.__class__.type_byte_length():
            raise ValueError("value must have equal byte length to coerce it")
        if isinstance(v, bytes):
            return cls.decode_bytes(v)
        return cls(v)

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        return cls(int.from_bytes(bytez, byteorder='little'))

    def encode_bytes(self) -> bytes:
        return self.to_bytes(length=self.__class__.type_byte_length(), byteorder='little')

    @classmethod
    def type_repr(cls) -> str:
        return f"uint{cls.type_byte_length()*8}"


class uint8(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 1


class uint16(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 2


class uint32(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 4


class uint64(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 8


class uint128(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 16


class uint256(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 32


class bit(boolean):
    pass


class byte(uint8):
    pass



from typing import Callable, NewType, List, Optional, Protocol, TypeVar
from hashlib import sha256


# Get the depth required for a given element count
# (in out): (0 0), (1 0), (2 1), (3 2), (4 2), (5 3), (6 3), (7 3), (8 3), (9 4)
def get_depth(elem_count: int) -> int:
    if elem_count <= 1:
        return 0
    return (elem_count - 1).bit_length()


Gindex = NewType("Gindex", int)

ROOT_GINDEX = Gindex(1)
LEFT_GINDEX = Gindex(2)
RIGHT_GINDEX = Gindex(3)


def to_gindex(index: int, depth: int):
    anchor = 1 << depth
    if index >= anchor:
        raise Exception("index %d too large for depth %d" % (index, depth))
    return anchor | index


def get_anchor_gindex(gindex: Gindex) -> Gindex:
    # noinspection PyTypeChecker
    return 1 << (gindex.bit_length() - 1)


Root = NewType("Root", bytes)

MerkleFn = NewType("MerkleFn", Callable[[Root, Root], Root])

ZERO_ROOT: Root = Root(b'\x00' * 32)


def merkle_hash(left: Root, right: Root) -> Root:
    return sha256(left + right).digest()


Link = Callable[["Node"], "Node"]
SummaryLink = Callable[[], "Node"]


class Node(Protocol):

    def getter(self, target: Gindex) -> "Node":
        raise NavigationError

    def setter(self, target: Gindex, expand: bool = False) -> Link:
        ...

    def summarize_into(self, target: Gindex) -> SummaryLink:
        setter = self.setter(target)
        getter = self.getter(target)
        return lambda: setter(RootNode(getter.merkle_root(merkle_hash)))

    def merkle_root(self, h: MerkleFn) -> Root:
        ...


# hashes of hashes of zeroes etc.
zero_hashes: List[Root] = [ZERO_ROOT]

for i in range(100):
    zero_hashes.append(merkle_hash(zero_hashes[i], zero_hashes[i]))


def zero_node(depth: int) -> "RootNode":
    return RootNode(zero_hashes[depth])


def identity(v: Node) -> Node:
    return v


def compose(inner: Link, outer: Link) -> Link:
    return lambda v: outer(inner(v))


class NavigationError(RuntimeError):
    pass


V = TypeVar('V', bound=Node)


class PairNode(Node):
    left: Node
    right: Node
    root: Optional[Root]

    def __init__(self, left: Node, right: Node):
        self.left = left
        self.right = right
        self.root = None

    def getter(self, target: Gindex) -> Node:
        if target < 1:
            raise NavigationError
        if target == 1:
            return self
        if target == 2:
            return self.left
        if target == 3:
            return self.right
        anchor = get_anchor_gindex(target)
        pivot = anchor >> 1
        unanchor = target ^ anchor
        if unanchor < pivot:
            return self.left.getter(Gindex(unanchor | pivot))
        else:
            return self.right.getter(Gindex(unanchor))

    def setter(self, target: Gindex, expand: bool = False) -> Link:
        if target < 1:
            raise NavigationError
        if target == 1:
            return identity
        if target == 2:
            return self.rebind_left
        if target == 3:
            return self.rebind_right
        anchor = get_anchor_gindex(target)
        pivot = anchor >> 1
        unanchor = target ^ anchor
        if unanchor < pivot:
            inner = self.left.setter(Gindex(unanchor | pivot), expand=expand)
            return compose(inner, self.rebind_left)
        else:
            inner = self.right.setter(Gindex(unanchor), expand=expand)
            return compose(inner, self.rebind_right)

    def rebind_left(self, v: Node) -> "PairNode":
        return PairNode(v, self.right)

    def rebind_right(self, v: Node) -> "PairNode":
        return PairNode(self.left, v)

    def merkle_root(self, h: MerkleFn) -> Root:
        if self.root is not None:
            return self.root
        self.root = h(self.left.merkle_root(h), self.right.merkle_root(h))
        return self.root

    def __repr__(self) -> str:
        return f"H({self.left}, {self.right})"


def subtree_fill_to_depth(bottom: Node, depth: int) -> Node:
    node = bottom
    while depth > 0:
        node = PairNode(node, node)
        depth -= 1
    return node


def subtree_fill_to_length(bottom: Node, depth: int, length: int) -> Node:
    if length > (1 << depth):
        raise Exception("too many nodes")
    if length == (1 << depth):
        return subtree_fill_to_depth(bottom, depth)
    if depth == 0:
        if length == 1:
            return bottom
        else:
            raise NavigationError
    if depth == 1:
        return PairNode(bottom, bottom if length > 1 else zero_node(0))
    else:
        anchor = 1 << depth
        pivot = anchor >> 1
        if length <= pivot:
            return PairNode(subtree_fill_to_length(bottom, depth - 1, length), zero_node(depth))
        else:
            return PairNode(
                subtree_fill_to_depth(bottom, depth-1),
                subtree_fill_to_length(bottom, depth-1, length - pivot)
            )


def subtree_fill_to_contents(nodes: List[Node], depth: int) -> Node:
    if len(nodes) > (1 << depth):
        raise Exception("too many nodes")
    if depth == 0:
        if len(nodes) == 1:
            return nodes[0]
        else:
            raise NavigationError
    if depth == 1:
        return PairNode(nodes[0], nodes[1] if len(nodes) > 1 else zero_node(0))
    else:
        anchor = 1 << depth
        pivot = anchor >> 1
        if len(nodes) <= pivot:
            return PairNode(subtree_fill_to_contents(nodes, depth - 1), zero_node(depth - 1))
        else:
            return PairNode(
                subtree_fill_to_contents(nodes[:pivot], depth-1),
                subtree_fill_to_contents(nodes[pivot:], depth-1)
            )


class RootNode(Node):
    root: Root

    def __init__(self, root: Root):
        self.root = root

    def getter(self, target: Gindex) -> Node:
        if target != 1:
            raise NavigationError
        return self

    def setter(self, target: Gindex, expand: bool = False) -> Link:
        if target < 1:
            raise NavigationError
        if target == 1:
            return identity
        if expand:
            child = zero_node(target.bit_length() - 2)
            return PairNode(child, child).setter(target, expand=True)
        else:
            raise NavigationError

    def merkle_root(self, h: MerkleFn) -> Root:
        return self.root

    def __repr__(self):
        return f"0x{self.root.hex()}"


def must_leaf(n: Node) -> Root:
    if not isinstance(n, RootNode):
        raise Exception(f"node {n} is not a rootnode")
    else:
        return n.root



# flake8:noqa F401  Ignore unused imports. Tests are a work in progress.
import pytest

from remerkleable.core import View, TypeDef, BasicView
from remerkleable.complex import Container, Vector, List
from remerkleable.basic import boolean, bit, uint, byte, uint8, uint16, uint32, uint64, uint128, uint256
from remerkleable.bitfields import Bitvector, Bitlist
from remerkleable.byte_vector import ByteVector, Bytes1, Bytes4, Bytes8, Bytes32, Bytes48, Bytes96
from remerkleable.core import BasicView, View, TypeDef
from remerkleable.tree import get_depth


def expect_value_error(fn, msg):
    try:
        fn()
        raise AssertionError(msg)
    except ValueError:
        pass


def test_subclasses():
    for u in [uint8, uint16, uint32, uint64, uint128, uint256]:
        assert issubclass(u, int)
        assert issubclass(u, View)
        assert issubclass(u, BasicView)
        assert isinstance(u, TypeDef)
    assert issubclass(boolean, BasicView)
    assert issubclass(boolean, View)
    assert isinstance(boolean, TypeDef)

    for c in [Container, List, Vector, Bytes32]:
        assert issubclass(c, View)
        assert isinstance(c, TypeDef)


def test_basic_instances():
    for u in [uint8, byte, uint16, uint32, uint64, uint128, uint256]:
        v = u(123)
        assert isinstance(v, int)
        assert isinstance(v, BasicView)
        assert isinstance(v, View)

    assert isinstance(boolean(True), BasicView)
    assert isinstance(boolean(False), BasicView)
    assert isinstance(bit(True), boolean)
    assert isinstance(bit(False), boolean)


def test_basic_value_bounds():
    max = {
        boolean: 2 ** 1,
        bit: 2 ** 1,
        uint8: 2 ** (8 * 1),
        byte: 2 ** (8 * 1),
        uint16: 2 ** (8 * 2),
        uint32: 2 ** (8 * 4),
        uint64: 2 ** (8 * 8),
        uint128: 2 ** (8 * 16),
        uint256: 2 ** (8 * 32),
    }
    for k, v in max.items():
        # this should work
        assert k(v - 1) == v - 1
        # but we do not allow overflows
        expect_value_error(lambda: k(v), f"no overflows allowed: type: {k}, value: {v}")

    for k, _ in max.items():
        # this should work
        assert k(0) == 0
        # but we do not allow underflows
        expect_value_error(lambda: k(-1), "no underflows allowed")


def test_container():
    class Foo(Container):
        a: uint8
        b: uint32

    empty = Foo()
    assert empty.a == uint8(0)
    assert empty.b == uint32(0)

    assert issubclass(Foo, Container)
    assert issubclass(Foo, View)

    assert Foo.is_fixed_byte_length()
    x = Foo(a=uint8(123), b=uint32(45))
    assert x.a == 123
    assert x.b == 45
    assert isinstance(x.a, uint8)
    assert isinstance(x.b, uint32)
    assert x.__class__.is_fixed_byte_length()

    class Bar(Container):
        a: uint8
        b: List[uint8, 1024]

    assert not Bar.is_fixed_byte_length()

    y = Bar(a=123, b=List[uint8, 1024](uint8(1), uint8(2)))
    assert y.a == 123
    assert isinstance(y.a, uint8)
    assert len(y.b) == 2
    assert isinstance(y.a, uint8)
    assert not y.__class__.is_fixed_byte_length()
    assert y.b[0] == 1
    v: List = y.b
    assert v.__class__.element_cls() == uint8
    assert v.__class__.limit() == 1024

    y.a = 42
    try:
        y.a = 256  # out of bounds
        assert False
    except ValueError:
        pass

    try:
        y.a = uint16(255)  # within bounds, wrong type
        assert False
    except ValueError:
        pass

    try:
        y.not_here = 5
        assert False
    except AttributeError:
        pass


def test_list():
    typ = List[uint64, 128]
    assert issubclass(typ, List)
    assert issubclass(typ, View)
    assert isinstance(typ, TypeDef)

    assert not typ.is_fixed_byte_length()

    assert len(typ()) == 0  # empty
    assert len(typ(uint64(0))) == 1  # single arg
    assert len(typ(uint64(i) for i in range(10))) == 10  # generator
    assert len(typ(uint64(0), uint64(1), uint64(2))) == 3  # args
    assert isinstance(typ(1, 2, 3, 4, 5)[4], uint64)  # coercion
    assert isinstance(typ(i for i in range(10))[9], uint64)  # coercion in generator

    v = typ(uint64(2), uint64(1))
    v[0] = uint64(123)
    assert v[0] == 123
    assert isinstance(v[0], uint64)

    assert isinstance(v, List)
    assert isinstance(v, View)

    assert len(typ([i for i in range(10)])) == 10  # cast py list to SSZ list

    foo = List[uint32, 128](0 for i in range(128))
    foo[0] = 123
    foo[1] = 654
    foo[127] = 222
    assert sum(foo) == 999
    try:
        foo[3] = 2 ** 32  # out of bounds
    except ValueError:
        pass

    try:
        foo[3] = uint64(2 ** 32 - 1)  # within bounds, wrong type
        assert False
    except ValueError:
        pass

    try:
        foo[128] = 100
        assert False
    except IndexError:
        pass

    try:
        foo[-1] = 100  # valid in normal python lists
        assert False
    except IndexError:
        pass

    try:
        foo[128] = 100  # out of bounds
        assert False
    except IndexError:
        pass


def test_bytesn_subclass():

    class Root(Bytes32):
        pass

    assert isinstance(Root(b'\xab' * 32), Bytes32)
    assert not isinstance(Root(b'\xab' * 32), Bytes48)
    assert issubclass(Root(b'\xab' * 32).__class__, Bytes32)
    assert issubclass(Root, Bytes32)

    assert not issubclass(Bytes48, Bytes32)

    assert len(Bytes32() + Bytes48()) == 80


def test_uint_math():
    assert uint8(0) + uint8(uint32(16)) == uint8(16)  # allow explicit casting to make invalid addition valid

    expect_value_error(lambda: uint8(0) - uint8(1), "no underflows allowed")
    expect_value_error(lambda: uint8(1) + uint8(255), "no overflows allowed")
    expect_value_error(lambda: uint8(0) + 256, "no overflows allowed")
    expect_value_error(lambda: uint8(42) + uint32(123), "no mixed types")
    expect_value_error(lambda: uint32(42) + uint8(123), "no mixed types")

    assert type(uint32(1234) + 56) == uint32


def test_container_depth():
    class SingleField(Container):
        foo: uint32
    assert SingleField.tree_depth() == 0

    class TwoField(Container):
        foo: uint32
        bar: uint64
    assert TwoField.tree_depth() == 1

    class ThreeField(Container):
        foo: uint32
        bar: uint64
        quix: uint8
    assert ThreeField.tree_depth() == 2

    class FourField(Container):
        foo: uint32
        bar: uint64
        quix: uint8
        more: uint32
    assert FourField.tree_depth() == 2

    class FiveField(Container):
        foo: uint32
        bar: uint64
        quix: uint8
        more: uint32
        fiv: uint8
    assert FiveField.tree_depth() == 3


@pytest.mark.parametrize("count, depth", [
    (0, 0), (1, 0), (2, 1), (3, 2), (4, 2), (5, 3), (6, 3), (7, 3), (8, 3), (9, 4)])
def test_tree_depth(count: int, depth: int):
    assert get_depth(count) == depth
